import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns

from .config import Configuration
from .genetic import generator, fitness, mutation, check_remaining_ticks, individual_to_melody
from deap import base, creator, tools, algorithms

from output import write_file


def ea_simple_with_elitism(population, toolbox, cxpb, mutpb, ngen, stats=None,
                           hall_of_fame=None, verbose=False):
    """This algorithm is similar to DEAP eaSimple() algorithm, with the modification that
    halloffame is used to implement an elitism mechanism. The individuals contained in the
    halloffame are directly injected into the next generation and are not subject to the
    genetic operators of selection, crossover and mutation.
    """
    logbook = tools.Logbook()
    logbook.header = ['gen', 'nevals'] + (stats.fields if stats else [])

    # Evaluate the individuals with an invalid fitness
    invalid_ind = [ind for ind in population if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit

    if hall_of_fame is None:
        raise ValueError("halloffame parameter must not be empty!")

    hall_of_fame.update(population)
    hof_size = len(hall_of_fame.items) if hall_of_fame.items else 0

    record = stats.compile(population) if stats else {}
    logbook.record(gen=0, nevals=len(invalid_ind), **record)
    if verbose:
        print(logbook.stream)

    # Begin the generational process
    for gen in range(1, ngen + 1):

        # Select the next generation individuals
        offspring = toolbox.select(population, len(population) - hof_size)

        # Vary the pool of individuals
        offspring = algorithms.varAnd(offspring, toolbox, cxpb, mutpb)

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # add the best back to population:
        offspring.extend(hall_of_fame.items)

        # Update the hall of fame with the generated individuals
        hall_of_fame.update(offspring)

        # Replace the current population by the offspring
        population[:] = offspring

        # Append the current generation statistics to the logbook
        record = stats.compile(population) if stats else {}
        logbook.record(gen=gen, nevals=len(invalid_ind), **record)
        if verbose:
            print(logbook.stream)

    return population, logbook


if __name__ == '__main__':
    toolbox = base.Toolbox()
    configuration = Configuration()
    gene_generator = generator(Configuration())
    toolbox.register("Note", gene_generator)
    creator.create("FitnessComposed", base.Fitness, weights=(-1.0, -1.0))
    creator.create("Individual", list, fitness=creator.FitnessComposed)
    toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.Note, n=32)
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("evaluate", fitness)
    toolbox.register("select", tools.selStochasticUniversalSampling)
    toolbox.register("mutate", mutation, configuration)
    toolbox.register("mate", tools.cxOnePoint)
    toolbox.register("select", tools.selTournament, tournsize=3)
    decorator = check_remaining_ticks()
    toolbox.decorate("mate", decorator)
    toolbox.decorate("mutate", decorator)
    population = toolbox.population(30)
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("min", np.min)
    stats.register("avg", np.mean)
    stats.register("min_axis", np.min, axis=0)
    stats.register("avg_axis", np.mean, axis=0)

    hof = tools.HallOfFame(10)
    population, logbook = ea_simple_with_elitism(population, toolbox, cxpb=0.4, mutpb=0.2,
                                                 ngen=100, stats=stats, hall_of_fame=hof)

    hof.update(population)
    best = hof.items[0]
    print("-- Best Ever Individual = ", best)
    print("-- Best Ever Fitness = ", best.fitness.values)

    # plot statistics:
    minFitnessValues, meanFitnessValues = logbook.select("min", "avg")
    plt.figure(1)
    sns.set_style("whitegrid")
    plt.xlabel('Generation')
    plt.ylabel('Min / Average Fitness')
    plt.title('Min and Average fitness over Generations')
    plt.plot(minFitnessValues, color="red")
    plt.plot(meanFitnessValues, color='green')

    minFitnessValues, meanFitnessValues = logbook.select("min_axis", "avg_axis")
    minSyncopationValues = list(map(lambda x: x[0], minFitnessValues))
    meanSyncopationValues = list(map(lambda x: x[0], meanFitnessValues))
    minNotesNo = list(map(lambda x: x[1], minFitnessValues))
    meanNotesNo = list(map(lambda x: x[1], minFitnessValues))

    plt.figure(2)
    plt.title('Syncopation min and avg')
    plt.xlabel('Generation')
    plt.ylabel('Min / Average syncopation')
    plt.plot(minSyncopationValues, color="red")
    plt.plot(meanSyncopationValues, color="green")
    plt.figure(3)
    plt.title('Notes min and avg')
    plt.xlabel('Generation')
    plt.ylabel('Min / Average notes')
    plt.plot(minNotesNo, color="red")
    plt.plot(meanNotesNo, color="green")
    plt.show()

    write_file(individual_to_melody(best), "best")
    write_file(individual_to_melody(best), 'best')



from itertools import takewhile, compress, accumulate
from typing import Tuple

from .config import Configuration, BEATS_PER_BAR
from .note import Note
from dataclasses import dataclass
import random

from .syncopation import weighted_note_to_beat, density


@dataclass
class Gene:
    pitch: int
    velocity: int
    remaining_ticks: int


@dataclass
class Individual:
    notes: list[Note]
    fitness: tuple[float]


Population = list[Individual]


def random_individual(config: Configuration):
    durations = config.rate
    start = 0
    notes = []

    while start < config.bars * BEATS_PER_BAR:
        key = random.choice(list(config.scale.notes)[30:40])
        velocity = random.randint(80, 100)
        end = min(start + random.choice(durations) * BEATS_PER_BAR, config.bars * BEATS_PER_BAR)
        notes.append(Note(key, velocity, start, end))
        start = end

    return Individual(notes, fitness(notes))


def generate_population(config: Configuration):
    return [random_individual(config) for i in range(40)]


# def mutation(individual: Individual, config: Configuration) -> Individual:
#     new_notes = []
#     for note in individual.notes:
#         change = random.random()
#         if change > 0.2:
#             new_end = min(note.start + random.choice(config.rate) * BEATS_PER_BAR, BEATS_PER_BAR * config.bars)
#             new_start = note.start
#             if len(new_notes) > 0 and not config.is_polyphonic and new_notes[-1].end > new_start:
#                 new_start = new_notes[-1].end
#             if new_start == new_end:
#                 continue
#             new_notes.append(Note(note.pitch, note.velocity, new_start, new_end))
#         else:
#             new_notes.append(Note(note.pitch, note.velocity, note.start, note.end))
#
#     return Individual(new_notes, fitness(new_notes))


def single_point_crossover(population: Population, config: Configuration) -> Population:
    max_notes = config.bars * (1 / min(config.rate)) * BEATS_PER_BAR
    for a, b in zip(population[::2], population[1::2]):
        change = random.random()
        if change > 0.4:
            cut = random.randint(0, max_notes) * min(config.rate)
            child = list(takewhile(lambda note: note.start < cut, a.notes)) + \
                list(takewhile(lambda note: note.start > cut, b.notes[::-1]))[::-1]
            population.append(Individual(child, fitness(child)))

    return population


def uniform_crossover(population: Population, config: Configuration) -> Population:
    max_notes = config.bars * (1 / min(config.rate)) * BEATS_PER_BAR
    for a, b in zip(population[::2], population[1::2]):
        change = random.random()
        if change > 0.4:
            part_a = [random.randint(0, 1)] * max_notes
            part_b = [0 if x == 1 else 1 for x in part_a]
            ...


def roulette_wheel(population: Population, config: Configuration) -> Population:
    new_population = []
    for i in range(20):
        probabilities = [0] + list(accumulate(map(lambda individual: individual.fitness, population)))
        change = random.uniform(0, probabilities[-1])
        index = len(list(takewhile(lambda x: x < change, probabilities)))
        new_population.append(population[index])

    return new_population


def individual_to_melody(individual: list[Gene]) -> list[Note]:
    notes = []
    start = 0.0
    time = 0.0
    for i, gene in enumerate(individual):
        time += 0.5
        if gene.remaining_ticks > 1:
            continue

        if gene.remaining_ticks == 1:
            notes.append(Note(gene.pitch, gene.velocity, start, time))
            start = time

        if gene.remaining_ticks == 0:
            start = time

    return notes


def fitness(genes: list[Gene]) -> tuple[float, float]:
    notes = individual_to_melody(genes)
    return abs(weighted_note_to_beat(notes) - 0.35), abs(density(genes) - 0.8)


def generator(config: Configuration):
    prev_gene: Gene | None = None

    def create_random_gene() -> Gene:
        nonlocal prev_gene
        if prev_gene is not None and prev_gene.remaining_ticks > 1:
            prev_gene = Gene(prev_gene.pitch, prev_gene.velocity, prev_gene.remaining_ticks - 1)
            return prev_gene

        tick = min(config.rate)
        duration = int(random.choice(config.rate) / tick)
        key = random.choice(list(config.scale.notes)[30:40])
        velocity = random.randint(80, 100)
        prev_gene = Gene(key, velocity, duration)
        return prev_gene

    return create_random_gene


def mutation(config: Configuration, genes: list[Gene]):
    for i, gene in enumerate(genes):
        change = random.random()
        if change > 0.1:
            gene.pitch = random.choice(list(config.scale.notes)[30:40])
        change = random.random()
        if change > 0.05 and gene.remaining_ticks == 1:
            gene.remaining_ticks = 0
    return genes,


def check_remaining_ticks():
    def decorator(func):
        def wrapper(*args, **kwargs):
            offspring = func(*args, **kwargs)
            for genes in offspring:
                for i, gene in enumerate(genes):
                    if gene.remaining_ticks == 0:
                        continue
                    gene.remaining_ticks = len(list(takewhile(lambda x: x.pitch == gene.pitch, genes[i:])))
            return offspring
        return wrapper
    return decorator



from dataclasses import field
from .note import *

LOOP_WAIT = 1.0 / 200  # 200 Hz
BEATS_PER_BAR = 4
MIDI_EVENT_NOTE_ON = 0x90


@dataclass
class Configuration:
    bars: int = 4
    density: float = 0.4
    rate: list[float] = field(default_factory=lambda: [1 / 4, 1 / 8])
    syncopation: float = 0.3
    scale: Scale = Scale(Key.C, Mode.MINOR)
    is_polyphonic: bool = False



from simplecoremidi import MIDISource
from midiutil import MIDIFile
from time import sleep
from .config import MIDI_EVENT_NOTE_ON
from .note import Note


def play(notes: list[Note]):
    source = MIDISource("Python")

    for note in notes:
        source.send((MIDI_EVENT_NOTE_ON, note.pitch, note.velocity))
        sleep(note.end - note.start)
        source.send((MIDI_EVENT_NOTE_ON, note.pitch, 0))


def write_file(notes: list[Note], file_name: str):
    midi = MIDIFile(1)
    midi.addTempo(0, 0, 120)

    for note in notes:
        midi.addNote(0, 0, note.pitch, note.start, note.end - note.start, note.velocity)

    with open(f'../midi/{file_name}.mid', "wb") as output_file:
        midi.writeFile(output_file)



from midi_generator.note import *
from midi_generator.config import *
from midi_generator.genetic import *
from midi_generator.output import *
from midi_generator.syncopation import *
from cli import print_cli


from .note import Note
from math import floor


def weighted_note_to_beat(notes: list[Note]) -> float:
    if len(notes) < 5:
        return -1
    total = 0

    for note in notes:
        left = floor(note.start)
        right = left + 1
        distance = min(note.start - left, abs(note.start - right))

        if distance == 0:
            continue
        if right < note.end < right + 1:
            total += 2 / distance
        else:
            total += 1 / distance

    return total / len(notes)


def density(notes) -> float:
    return len(list(filter(lambda note: note.remaining_ticks != 0, notes))) / len(notes)
